/*
Wherever '<something>--extends' keys are discovered in piped files - the specified file is included and merged with the following object.
Looks for the EXTENDS in the JSON.
Where if finds it - it merges the objects.

Based on https://github.com/joshswan/gulp-merge-json

TODO:
Enhance that the extensions can be json or yaml. Now they must be YAML which is wierd.
If they are yaml - they are converted to json.

Support relative paths instead of only absolute project paths as now.

Merge a target element in the target document - not the entire thing.
*/

'use strict';

var gutil = require('gulp-util');
var merge = require('deepmerge');
var through = require('through2');
var util = require('util');
var yamljs = require('yamljs');

var PLUGIN_NAME = 'gulp-json-include-and-merge';

var INCLUDE_AND_MERGE_KEY = "--extends";

module.exports = function(startObj, endObj) {
  if ((startObj && typeof startObj !== 'object') || (endObj && typeof endObj !== 'object')) {
    throw new gutil.PluginError(PLUGIN_NAME, PLUGIN_NAME + ': Invalid start and/or end object!');
  }

  function logObject(name,obj){
    console.log('>');
    console.log(name + ":");
    console.log(util.inspect(obj, {showHidden: false, depth: null}));
    console.log('>');
  }

  /**
  Private.
  Walks object recursively and merges when it finds the 'search' string.
  */
  function walkObjectAndMerge(obj, objParent, search) {
    //logObject("walkObject:", obj);
    var objects = [];
    for (var key in obj) {
        if (!obj.hasOwnProperty(key)) continue;
        console.log(key);
        var value = obj[key];
        if (typeof obj[key] == 'object') {
            objects = objects.concat(walkObjectAndMerge(obj[key], obj, search));
        } else {
            //property.
            if (key.indexOf(search) > -1) {
              console.log("Found EXTENDS!");
              //logObject("parent:", objParent);
              //logObject("obj:", obj);

              var where = key.indexOf(search);
              var keyToExtend = key.substring(0,where);
              //console.log("key to extend:" + keyToExtend);
              //var where = value.indexOf(search);
              //var a =
              //var path = value.substr(where + search.length);
              var path = value;
              //console.log("path:" + path);

              if (path){
                // remove the extends from the object.
                delete obj[key];

                var extended = yamljs.load(path);
                //logObject("extended", extended);

                //logObject("the extention", obj);
                //objects.push(obj);
                // merge the current object.
                var merged = merge(extended, obj);
                //logObject("extended node", merged);
                obj[keyToExtend] = merged[keyToExtend];
                //logObject("obj", obj);
                //logObject("objParent", objParent);
              }
            }
        }
    }
    //return objects;
}

  return through.obj(function(file, enc, callback) {

    if (file.isStream()) {
      return this.emit('error', new gutil.PluginError(PLUGIN_NAME, PLUGIN_NAME + ': Streaming not supported!'));
    }

    if (file.isBuffer()) {

      if (file.contents.length === 0) {
        this.emit('error', new PluginError(PLUGIN_NAME, 'File ' + file.path +
            ' is empty. YAML loader cannot load empty content'));
        return callback();
      }

      try {
        var json = file.contents.toString('utf8');
        var fileObj = JSON.parse(json);

        //Find extends
        walkObjectAndMerge(fileObj, fileObj, INCLUDE_AND_MERGE_KEY);

        logObject("fileObj", fileObj);

        var mergedBuffer = new Buffer(JSON.stringify(fileObj, null, '  '),'utf8');
        file.contents = mergedBuffer;

      } catch (err) {
       return this.emit('error', new gutil.PluginError(PLUGIN_NAME, err));
      }

      this.push(file);
      callback();
    }

  });
};
